# -*- coding: utf-8 -*-

from __future__ import division, print_function

import re
import glob
import os
import sys
import unittest
import warnings
import shutil
from tempfile import mkstemp

from unittest import TestCase as BaseTestCase

from mutagen._compat import PY3
from mutagen._toolsutil import fsencoding, is_fsnative


DATA_DIR = os.path.join(os.path.dirname(os.path.realpath(__file__)), "data")
if os.name == "nt" and not PY3:
    DATA_DIR = DATA_DIR.decode("ascii")
assert is_fsnative(DATA_DIR)


if os.name != "nt":
    try:
        u"öäü".encode(fsencoding())
    except ValueError:
        raise RuntimeError("This test suite needs a unicode locale encoding. "
                           "Try setting LANG=C.UTF-8")


# Make sure we see all deprecation warnings so we either have to avoid them
# or capture them in the test suite
warnings.simplefilter("always")
warnings.simplefilter("ignore", PendingDeprecationWarning)


def get_temp_copy(path):
    """Returns a copy of the file with the same extension"""

    ext = os.path.splitext(path)[-1]
    fd, filename = mkstemp(suffix=ext)
    os.close(fd)
    shutil.copy(path, filename)
    return filename


def get_temp_empty(ext=""):
    """Returns an empty file with the extension"""

    fd, filename = mkstemp(suffix=ext)
    os.close(fd)
    return filename


class TestCase(BaseTestCase):

    def failUnlessRaisesRegexp(self, exc, re_, fun, *args, **kwargs):
        def wrapped(*args, **kwargs):
            try:
                fun(*args, **kwargs)
            except Exception as e:
                self.failUnless(re.search(re_, str(e)))
                raise
        self.failUnlessRaises(exc, wrapped, *args, **kwargs)

    # silence deprec warnings about useless renames
    failUnless = BaseTestCase.assertTrue
    failIf = BaseTestCase.assertFalse
    failUnlessEqual = BaseTestCase.assertEqual
    failUnlessRaises = BaseTestCase.assertRaises
    failUnlessAlmostEqual = BaseTestCase.assertAlmostEqual
    failIfEqual = BaseTestCase.assertNotEqual
    failIfAlmostEqual = BaseTestCase.assertNotAlmostEqual
    assertEquals = BaseTestCase.assertEqual
    assertNotEquals = BaseTestCase.assertNotEqual
    assert_ = BaseTestCase.assertTrue

    def assertReallyEqual(self, a, b):
        self.assertEqual(a, b)
        self.assertEqual(b, a)
        self.assertTrue(a == b)
        self.assertTrue(b == a)
        self.assertFalse(a != b)
        self.assertFalse(b != a)
        if not PY3:
            self.assertEqual(0, cmp(a, b))
            self.assertEqual(0, cmp(b, a))

    def assertReallyNotEqual(self, a, b):
        self.assertNotEqual(a, b)
        self.assertNotEqual(b, a)
        self.assertFalse(a == b)
        self.assertFalse(b == a)
        self.assertTrue(a != b)
        self.assertTrue(b != a)
        if not PY3:
            self.assertNotEqual(0, cmp(a, b))
            self.assertNotEqual(0, cmp(b, a))


def import_tests():
    tests = []

    for name in glob.glob(
            os.path.join(os.path.dirname(__file__), "test_*.py")):
        module_name = "tests." + os.path.basename(name)
        mod = __import__(module_name[:-3], {}, {}, [])
        mod = getattr(mod, os.path.basename(name)[:-3])

        tests.extend(get_tests_from_mod(mod))

    return list(set(tests))


def get_tests_from_mod(mod):
    tests = []
    for name in dir(mod):
        obj = getattr(mod, name)
        if isinstance(obj, type) and issubclass(obj, BaseTestCase) and \
                obj is not TestCase:
            tests.append(obj)
    return tests


class Result(unittest.TestResult):

    separator1 = '=' * 70
    separator2 = '-' * 70

    def addSuccess(self, test):
        unittest.TestResult.addSuccess(self, test)
        sys.stdout.write('.')

    def addError(self, test, err):
        unittest.TestResult.addError(self, test, err)
        sys.stdout.write('E')

    def addFailure(self, test, err):
        unittest.TestResult.addFailure(self, test, err)
        sys.stdout.write('F')

    def printErrors(self):
        succ = self.testsRun - (len(self.errors) + len(self.failures))
        v = "%3d" % succ
        count = 50 - self.testsRun
        sys.stdout.write((" " * count) + v + "\n")
        self.printErrorList('ERROR', self.errors)
        self.printErrorList('FAIL', self.failures)

    def printErrorList(self, flavour, errors):
        for test, err in errors:
            sys.stdout.write(self.separator1 + "\n")
            sys.stdout.write("%s: %s\n" % (flavour, str(test)))
            sys.stdout.write(self.separator2 + "\n")
            sys.stdout.write("%s\n" % err)


class Runner(object):
    def run(self, test):
        suite = unittest.makeSuite(test)
        pref = '%s (%d): ' % (test.__name__, len(suite._tests))
        print (pref + " " * (25 - len(pref)), end="")
        result = Result()
        suite(result)
        result.printErrors()
        return bool(result.failures + result.errors)


def check():
    from tests.quality import test_pep8
    from tests.quality import test_pyflakes

    tests = get_tests_from_mod(test_pep8)
    tests += get_tests_from_mod(test_pyflakes)

    runner = Runner()
    failures = 0
    for test in sorted(tests, key=lambda c: c.__name__):
        failures += runner.run(test)

    return len(tests), failures


def unit(run=[], exitfirst=False):
    tests = import_tests()

    runner = Runner()
    failures = 0
    filtered = [t for t in tests if not run or t.__name__ in run]

    for test in sorted(filtered, key=lambda c: c.__name__):
        if failures and exitfirst:
            break

        failures += runner.run(test)

    return len(filtered), failures



# -*- coding: utf-8 -*-

# Copyright 2015 Christoph Reiter
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.

import os
import sys
import signal
import locale
import contextlib
import optparse
import ctypes

from ._compat import text_type, PY2, PY3, iterbytes


def split_escape(string, sep, maxsplit=None, escape_char="\\"):
    """Like unicode/str/bytes.split but allows for the separator to be escaped

    If passed unicode/str/bytes will only return list of unicode/str/bytes.
    """

    assert len(sep) == 1
    assert len(escape_char) == 1

    if isinstance(string, bytes):
        if isinstance(escape_char, text_type):
            escape_char = escape_char.encode("ascii")
        iter_ = iterbytes
    else:
        iter_ = iter

    if maxsplit is None:
        maxsplit = len(string)

    empty = string[:0]
    result = []
    current = empty
    escaped = False
    for char in iter_(string):
        if escaped:
            if char != escape_char and char != sep:
                current += escape_char
            current += char
            escaped = False
        else:
            if char == escape_char:
                escaped = True
            elif char == sep and len(result) < maxsplit:
                result.append(current)
                current = empty
            else:
                current += char
    result.append(current)
    return result


class SignalHandler(object):

    def __init__(self):
        self._interrupted = False
        self._nosig = False
        self._init = False

    def init(self):
        signal.signal(signal.SIGINT, self._handler)
        signal.signal(signal.SIGTERM, self._handler)
        if os.name != "nt":
            signal.signal(signal.SIGHUP, self._handler)

    def _handler(self, signum, frame):
        self._interrupted = True
        if not self._nosig:
            raise SystemExit("Aborted...")

    @contextlib.contextmanager
    def block(self):
        """While this context manager is active any signals for aborting
        the process will be queued and exit the program once the context
        is left.
        """

        self._nosig = True
        yield
        self._nosig = False
        if self._interrupted:
            raise SystemExit("Aborted...")


def get_win32_unicode_argv():
    """Returns a unicode argv under Windows and standard sys.argv otherwise"""

    if os.name != "nt" or not PY2:
        return sys.argv

    import ctypes
    from ctypes import cdll, windll, wintypes

    GetCommandLineW = cdll.kernel32.GetCommandLineW
    GetCommandLineW.argtypes = []
    GetCommandLineW.restype = wintypes.LPCWSTR

    CommandLineToArgvW = windll.shell32.CommandLineToArgvW
    CommandLineToArgvW.argtypes = [
        wintypes.LPCWSTR, ctypes.POINTER(ctypes.c_int)]
    CommandLineToArgvW.restype = ctypes.POINTER(wintypes.LPWSTR)

    LocalFree = windll.kernel32.LocalFree
    LocalFree.argtypes = [wintypes.HLOCAL]
    LocalFree.restype = wintypes.HLOCAL

    argc = ctypes.c_int()
    argv = CommandLineToArgvW(GetCommandLineW(), ctypes.byref(argc))
    if not argv:
        return

    res = argv[max(0, argc.value - len(sys.argv)):argc.value]

    LocalFree(argv)

    return res


def fsencoding():
    """The encoding used for paths, argv, environ, stdout and stdin"""

    if os.name == "nt":
        return ""

    return locale.getpreferredencoding() or "utf-8"


def fsnative(text=u""):
    """Returns the passed text converted to the preferred path type
    for each platform.
    """

    assert isinstance(text, text_type)

    if os.name == "nt" or PY3:
        return text
    else:
        return text.encode(fsencoding(), "replace")
    return text


def is_fsnative(arg):
    """If the passed value is of the preferred path type for each platform.
    Note that on Python3+linux, paths can be bytes or str but this returns
    False for bytes there.
    """

    if PY3 or os.name == "nt":
        return isinstance(arg, text_type)
    else:
        return isinstance(arg, bytes)


def print_(*objects, **kwargs):
    """A print which supports bytes and str+surrogates under python3.

    Needed so we can print anything passed to us through argv and environ.
    Under Windows only text_type is allowed.

    Arguments:
        objects: one or more bytes/text
        linesep (bool): whether a line separator should be appended
        sep (bool): whether objects should be printed separated by spaces
    """

    linesep = kwargs.pop("linesep", True)
    sep = kwargs.pop("sep", True)
    file_ = kwargs.pop("file", None)
    if file_ is None:
        file_ = sys.stdout

    old_cp = None
    if os.name == "nt":
        # Try to force the output to cp65001 aka utf-8.
        # If that fails use the current one (most likely cp850, so
        # most of unicode will be replaced with '?')
        encoding = "utf-8"
        old_cp = ctypes.windll.kernel32.GetConsoleOutputCP()
        if ctypes.windll.kernel32.SetConsoleOutputCP(65001) == 0:
            encoding = getattr(sys.stdout, "encoding", None) or "utf-8"
            old_cp = None
    else:
        encoding = fsencoding()

    try:
        if linesep:
            objects = list(objects) + [os.linesep]

        parts = []
        for text in objects:
            if isinstance(text, text_type):
                if PY3:
                    try:
                        text = text.encode(encoding, 'surrogateescape')
                    except UnicodeEncodeError:
                        text = text.encode(encoding, 'replace')
                else:
                    text = text.encode(encoding, 'replace')
            parts.append(text)

        data = (b" " if sep else b"").join(parts)
        try:
            fileno = file_.fileno()
        except (AttributeError, OSError, ValueError):
            # for tests when stdout is replaced
            try:
                file_.write(data)
            except TypeError:
                file_.write(data.decode(encoding, "replace"))
        else:
            file_.flush()
            os.write(fileno, data)
    finally:
        # reset the code page to what we had before
        if old_cp is not None:
            ctypes.windll.kernel32.SetConsoleOutputCP(old_cp)


class OptionParser(optparse.OptionParser):
    """OptionParser subclass which supports printing Unicode under Windows"""

    def print_help(self, file=None):
        print_(self.format_help(), file=file)
